/*
 * Copyright 2009-2012 Yi-Qing WANG
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */


/**
 * @file SPLE_lib.h
 * @brief header file for SPLE denoising algorithm
 * @brief SPLE is implemented mainly with Eigen (Newmat for SVD)
 * @brief PLE is implemented mainly with Newmat
 * @author Yi-Qing WANG <yiqing.wang@polytechnique.edu>
 */

#ifndef SPLE_H
#define SPLE_H

#include <Eigen/Dense>
using namespace Eigen;

/**@brief present the patch array with matrices*/
void vectorArray2Matrix(
		VectorXd ** varray
,		int row_id
,		MatrixXd & vmatrix
);

/**
 * @brief reduce a gray image into patches
 *
 * @param image a matrix representing one image
 * @param overlap number of columns (resp. rows) two horizontal (resp. vectical) neighbor patches share
 * @param patch_size patch dimension: patch_size * patch_size
 * @return an array of vectorized patches
 */
VectorXd ** image2patches(
		MatrixXd const & image
,		int image_rows
,		int image_cols
,		int overlap
,		int patch_size
);

/** @brief number of patches required to cover the whole image by row or by column */
int num_patches(
		int n_pixels
,		int patch_size
,		int overlap
);

/**
 * @brief return the number of factors in each mixture component
 *
 * @param std_num_factors number of factors in an oriented model
 * @param data_size the number of pixels in a patch
 * @param model the current model
 * @param num_models the number of mixture components
 */
int ret_num_factors(
		int std_num_factors
,		int data_size
, 		int model
,		int num_models
);

/**
 * @brief read in the initial GFM setup
 *
 * @param prior the model priors (mixing weights)
 * @param mus model means
 * @param factors model factors
 */
void read_config(
		int data_size
,		int std_num_factors
,		int num_models
,		VectorXd *& prior
,		VectorXd *& mus
,		MatrixXd *& factors
);

/**
 * @brief a multivariate Gaussian density value
 *
 * @param x the observation
 * @param mu the Gaussian model mean
 * @param inv_eig_vals the inverse of the eigenvalues of the sum of factor + noise variance
 * @param eig_vecs the principal components of the subspace spanned by the factors
 * @param squared_sigma squared sigma
 * @param num_factors number of factors in the current model
 * @return density up to an absolute constant
 */
double gaussian_density_EM(
			VectorXd const & x
,			VectorXd const & mu
, 			VectorXd const & inv_eig_vals
,			MatrixXd const & eig_vecs
,			double squared_sigma
,			int num_factors
);

/**
 * @brief calculate the posterior probability for a patch to belong to a mixture component
 *
 * @param noisy_patches_at_coordinates noisy patch array
 * @param factors mixture component factors
 * @param mus model means
 * @param prior model priors
 * @param map_rows the number of patches covering the image by row
 * @param map_cols the number of patches covering the image by column
 * @param data_size the number of pixels in a patch
 * @param std_num_factors the number of factors in an oriented model
 * @param sigma the standard deviation of corrupting noise
 * @param responsibility the posterior probability for each patch in the array
 * @param patch_map a patch to model mapping in the form of an integer matrix
 */
void calc_responsibilities(
		VectorXd ** noisy_patches_at_coordinates
, 		MatrixXd * factors
, 		VectorXd * mus
,		VectorXd * prior
,		int map_rows
,		int map_cols
,		int data_size
,		int std_num_factors
,	 	double sigma
, 		VectorXd **& responsibility
, 		MatrixXi & patch_map
);

/**
 * @brief a wrapper of SVD implemented in NEWMAT
 *
 * @param cov_mat the matrix to SVD
 * @param std_num_factors the number of factors contained in an oriented mixture component
 * @param data_size the number of pixels in a patch
 * @param model the current component
 * @param num_models the number of mixture components
 * @param squared_sigma variance of corrupting noise
 * @param inverse_eig_vals whether to compute the inverse of eigenvalues for Wiener filtering for instance
 * @param trim whether to remove the factors attrbuted to noise rather than signal
 * @param eig_vecs returned eigenvectors of cov_mat
 * @param eig_vecs returned eigenvalues of cov_mat
 */
void deduce_eigens(
		MatrixXd & cov_mat
,		int std_num_factors
,		int data_size
,		int model
,		int num_models
,		double squared_sigma
,		bool inverse_eig_vals
,		bool trim
,		MatrixXd & eig_vecs
,		VectorXd & eig_vals
);

/**
 * @brief update the mixture model means and factors
 *
 * @param noisy_patches noisy patch array
 * @param responsibility the posterior probability for each noisy patch
 * @param prior model priors (or mixing weights)
 * @param num_models number of mixture components
 * @param map_rows the number of patches to cover the image by row
 * @param map_cols the number of patches to cover the image by column
 * @param sigma the standard deviation of corrupting noise
 * @param mus the model means
 * @param factors the mixture component factors
 */
void update_PPCA(
		VectorXd ** noisy_patches
,		VectorXd ** responsibility
,		VectorXd * prior
,		int num_models
,		int map_rows
,		int map_cols
,		double sigma
,		VectorXd *& mus
,		MatrixXd *& factors
);

/**
 * @brief write out as a PNG
 *
 * @param file_name the image name
 * @param num_channels RGB or Gray
 */
void imwrite(
		const char * file_name
, 		MatrixXd * image
, 		int num_channels
);

/** @brief print out an integer matrix as a color image */
void print_patch_map(
		MatrixXi const & patch_map
,		bool is_patch_map = true
);

#endif
